template <class T, const IID* piid, class CDV = CComDynamicUnkArray >
class ATL_NO_VTABLE IAppartSafeConnectionPointImpl : public _ICPLocator<piid>
{
	typedef CComEnum<IEnumConnections, &__uuidof(IEnumConnections), CONNECTDATA,
		_Copy<CONNECTDATA> > CComEnumConnections;
	typedef CDV _CDV;
public:
	IAppartSafeConnectionPointImpl();
	~IAppartSafeConnectionPointImpl();
	STDMETHOD(_LocCPQueryInterface)(REFIID riid, void ** ppvObject)
	{
#ifndef _ATL_OLEDB_CONFORMANCE_TESTS
		ATLASSERT(ppvObject != NULL);
#endif
		if (ppvObject == NULL)
			return E_POINTER;
		*ppvObject = NULL;

		if (InlineIsEqualGUID(riid, __uuidof(IConnectionPoint)) || InlineIsEqualUnknown(riid))
		{
			*ppvObject = this;
			AddRef();
#ifdef _ATL_DEBUG_INTERFACES
			_AtlDebugInterfacesModule.AddThunk((IUnknown**)ppvObject, _T("IAppartSafeConnectionPointImpl"), riid);
#endif // _ATL_DEBUG_INTERFACES
			return S_OK;
		}
		else
			return E_NOINTERFACE;
	}

	STDMETHOD(GetConnectionInterface)(IID* piid2)
	{
		if (piid2 == NULL)
			return E_POINTER;
		*piid2 = *piid;
		return S_OK;
	}
	STDMETHOD(GetConnectionPointContainer)(IConnectionPointContainer** ppCPC)
	{
		T* pT = static_cast<T*>(this);
		// No need to check ppCPC for NULL since QI will do that for us
		return pT->QueryInterface(__uuidof(IConnectionPointContainer), (void**)ppCPC);
	}
	STDMETHOD(Advise)(IUnknown* pUnkSink, DWORD* pdwCookie);
	STDMETHOD(Unadvise)(DWORD dwCookie);
	STDMETHOD(EnumConnections)(IEnumConnections** ppEnum);
	CDV 
		m_vec;
	std::map<DWORD,DWORD> 
		m_mapCookie2GIT;
	CComPtr<IGlobalInterfaceTable> 
		m_AppartGit;
};

template <class T, const IID* piid, class CDV>
IAppartSafeConnectionPointImpl<T, piid, CDV>::IAppartSafeConnectionPointImpl()
{
	HRESULT hr = m_AppartGit.CoCreateInstance(CLSID_StdGlobalInterfaceTable);
	ATLASSERT(!FAILED(hr));
}

template <class T, const IID* piid, class CDV>
IAppartSafeConnectionPointImpl<T, piid, CDV>::~IAppartSafeConnectionPointImpl()
{
	// cleanup unknowns
	IUnknown** pp = m_vec.begin();
	while (pp < m_vec.end())
	{
		if (*pp != NULL)
			(*pp)->Release();
		pp++;
	}
	// cleanup git's
	std::map<DWORD,DWORD>::iterator pIt;
	for (pIt=m_mapCookie2GIT.begin(); pIt!=m_mapCookie2GIT.end(); pIt++)
		m_AppartGit->RevokeInterfaceFromGlobal((*pIt).second);
}

template <class T, const IID* piid, class CDV>
STDMETHODIMP IAppartSafeConnectionPointImpl<T, piid, CDV>::Advise(IUnknown* pUnkSink,
	DWORD* pdwCookie)
{
	T* pT = static_cast<T*>(this);
	IUnknown* p;
	HRESULT hRes = S_OK;
	if (pdwCookie != NULL)
		*pdwCookie = 0;
	if (pUnkSink == NULL || pdwCookie == NULL)
		return E_POINTER;
	IID iid;
	GetConnectionInterface(&iid);
	hRes = pUnkSink->QueryInterface(iid, (void**)&p);
	if (SUCCEEDED(hRes))
	{
		pT->Lock();
		*pdwCookie = m_vec.Add(p);
		hRes = (*pdwCookie != NULL) ? S_OK : CONNECT_E_ADVISELIMIT;
		if (!FAILED(hRes))
		{
			DWORD dwGitCookie;
			hRes = m_AppartGit->RegisterInterfaceInGlobal(p,iid,&dwGitCookie);
			if (!FAILED(hRes))
				m_mapCookie2GIT[*pdwCookie]=dwGitCookie;
		}
		pT->Unlock();
		if (hRes != S_OK)
			p->Release();
	}
	else if (hRes == E_NOINTERFACE)
		hRes = CONNECT_E_CANNOTCONNECT;
	if (FAILED(hRes))
		*pdwCookie = 0;
	return hRes;
}

template <class T, const IID* piid, class CDV>
STDMETHODIMP IAppartSafeConnectionPointImpl<T, piid, CDV>::Unadvise(DWORD dwCookie)
{
	T* pT = static_cast<T*>(this);
	pT->Lock();
	IUnknown* p = m_vec.GetUnknown(dwCookie);
	HRESULT hRes = m_vec.Remove(dwCookie) ? S_OK : CONNECT_E_NOCONNECTION;
	if (m_mapCookie2GIT.count(dwCookie))
	{
		m_AppartGit->RevokeInterfaceFromGlobal(m_mapCookie2GIT[dwCookie]);
		m_mapCookie2GIT.erase(dwCookie);
	}
	pT->Unlock();
	if (hRes == S_OK && p != NULL)
		p->Release();
	return hRes;
}

template <class T, const IID* piid, class CDV>
STDMETHODIMP IAppartSafeConnectionPointImpl<T, piid, CDV>::EnumConnections(
	IEnumConnections** ppEnum)
{
	if (ppEnum == NULL)
		return E_POINTER;
	*ppEnum = NULL;
	CComObject<CComEnumConnections>* pEnum = NULL;
	ATLTRY(pEnum = new CComObject<CComEnumConnections>)
	if (pEnum == NULL)
		return E_OUTOFMEMORY;
	T* pT = static_cast<T*>(this);
	pT->Lock();
	CONNECTDATA* pcd = NULL;
	ATLTRY(pcd = new CONNECTDATA[m_vec.end()-m_vec.begin()])
	if (pcd == NULL)
	{
		delete pEnum;
		pT->Unlock();
		return E_OUTOFMEMORY;
	}
	CONNECTDATA* pend = pcd;
	// Copy the valid CONNECTDATA's
	std::map<DWORD,DWORD>::iterator pIt;
	for (pIt=m_mapCookie2GIT.begin(); pIt!=m_mapCookie2GIT.end(); pIt++)
	{
		IUnknown *pUnk;
		if (pT->m_bGitEvents)
			m_AppartGit->GetInterfaceFromGlobal((*pIt).second,*piid,(void **)&pUnk);
		else
		{
			pUnk = m_vec.GetUnknown((*pIt).first);
			if (pUnk) pUnk->AddRef();
		}

		if (pUnk != NULL)
		{
			pend->pUnk = pUnk;
			pend->dwCookie = (*pIt).first;
			pend++;
		}
	}
	// don't copy the data, but transfer ownership to it
	pEnum->Init(pcd, pend, NULL, AtlFlagTakeOwnership);
	pT->Unlock();
	HRESULT hRes = pEnum->_InternalQueryInterface(__uuidof(IEnumConnections), (void**)ppEnum);
	if (FAILED(hRes))
		delete pEnum;
	return hRes;
}
