// Copyright (C) 2006 Pushok Software http://www.pushok.com/soft_svncom.php
//
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html
//
// Subversion is Copyrght of(c) 2000-2004 CollabNet.  All rights reserved.
// Full license available here http://subversion.tigris.org/license-1.html.

#include "stdafx.h"
#include "SVNAuthProvider.h"
#include "Helpers.h"
#include "svn_config.h"

// CSVNAuthProvider

svn_error_t *
pcom__auth_simple_prompt (svn_auth_cred_simple_t **cred_p,
							void *baton,
							const char *realm,
							const char *username,
							svn_boolean_t may_save,
							apr_pool_t *pool)
{
	svn_auth_cred_simple_t *ret = (svn_auth_cred_simple_t *)apr_pcalloc (pool, sizeof (*ret));

	CComGITPtr<ISVNAuthProvider> gProvider;
	CComPtr<ISVNAuthProvider> oProvider;
	if (FAILED(gProvider.Attach(*(DWORD *)baton)))
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	if (FAILED(gProvider.CopyTo(&oProvider)))
	{
		gProvider.Detach();
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	}
	gProvider.Detach();
	oProvider->put_RealM(Utf82Ole(realm));
	oProvider->put_MaySave(may_save?VARIANT_TRUE:VARIANT_FALSE);

	CComBSTR bsUserName,bsUserPassword;

	if (username)
	{
		ret->username = apr_pstrdup (pool, username);
		oProvider->put_DefUserName(Utf82Ole(username));
		oProvider->FireAskEvent(SVNAuthProviderAskPassword,oProvider);
	}
	else
	{
		oProvider->FireAskEvent(SVNAuthProviderAskUserAndPassword,oProvider);
		oProvider->get_DefUserName(&bsUserName);
		ret->username = apr_pstrdup (pool, Ole2Utf8(bsUserName));
	}
	oProvider->get_DefPassword(&bsUserPassword);
	ret->password = apr_pstrdup(pool, Ole2Utf8(bsUserPassword));

	VARIANT_BOOL bMaySave;
	oProvider->get_MaySave(&bMaySave);
	ret->may_save = bMaySave==VARIANT_TRUE;

	*cred_p = ret;
	return SVN_NO_ERROR;
}

svn_error_t *
pcom__auth_username_prompt (svn_auth_cred_username_t **cred,
                            void *baton,
                            const char *realm,
                            svn_boolean_t may_save,
                            apr_pool_t *pool)
{
	svn_auth_cred_username_t *ret = (svn_auth_cred_username_t *)apr_pcalloc (pool, sizeof (*ret));

	CComGITPtr<ISVNAuthProvider> gProvider;
	CComPtr<ISVNAuthProvider> oProvider;
	if (FAILED(gProvider.Attach(*(DWORD *)baton)))
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	if (FAILED(gProvider.CopyTo(&oProvider)))
	{
		gProvider.Detach();
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	}
	gProvider.Detach();
	oProvider->put_RealM(Utf82Ole(realm));
	oProvider->put_MaySave(may_save?VARIANT_TRUE:VARIANT_FALSE);

	CComBSTR bsUserName;

	oProvider->FireAskEvent(SVNAuthProviderAskPassword,oProvider);
	oProvider->get_DefPassword(&bsUserName);
	ret->username = apr_pstrdup(pool, Ole2Utf8(bsUserName));

	VARIANT_BOOL bMaySave;
	oProvider->get_MaySave(&bMaySave);
	ret->may_save = bMaySave==VARIANT_TRUE;

	*cred = ret;
	return SVN_NO_ERROR;
}

/* This implements 'svn_auth_ssl_server_trust_prompt_func_t'. */
svn_error_t *
pcom__auth_ssl_server_trust_prompt (svn_auth_cred_ssl_server_trust_t **cred_p,
									void *baton,
									const char *realm,
									apr_uint32_t failures,
									const svn_auth_ssl_server_cert_info_t *cert_info,
									svn_boolean_t may_save,
									apr_pool_t *pool)
{
	CComGITPtr<ISVNAuthProvider> gProvider;
	CComPtr<ISVNAuthProvider> oProvider;
	if (FAILED(gProvider.Attach(*(DWORD *)baton)))
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	if (FAILED(gProvider.CopyTo(&oProvider)))
	{
		gProvider.Detach();
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	}
	gProvider.Detach();

	CString sBuf = "Error validating server certificate for ";
	sBuf += realm;
	sBuf += ":\n";

	if (failures & SVN_AUTH_SSL_UNKNOWNCA)
		sBuf += " - The certificate is not issued by a trusted authority. Use the\n"
			"   fingerprint to validate the certificate manually!\n";
	if (failures & SVN_AUTH_SSL_CNMISMATCH)
		sBuf += " - The certificate hostname does not match.\n";
	if (failures & SVN_AUTH_SSL_NOTYETVALID)
		sBuf += " - The certificate is not yet valid.\n";
	if (failures & SVN_AUTH_SSL_EXPIRED)
		sBuf += " - The certificate has expired.\n";
	if (failures & SVN_AUTH_SSL_OTHER)
		sBuf += " - The certificate has an unknown error.\n";

	CString msg = "Certificate information:\n - Hostname: ";
	msg += cert_info->hostname;
	msg += "\n - Valid: from ";
	msg += cert_info->valid_from;
	msg += " until ";
	msg += cert_info->valid_until;
	msg += "\n - Issuer: ";
	msg += cert_info->issuer_dname;
	msg += "\n - Fingerprint: ";
	msg += cert_info->fingerprint;
	msg += "\n";
	sBuf += msg;

	if (may_save)
			sBuf += "\"Cansel\"-Reject, accept \"Yes\"-temporarily or accept \"No\"-permanently? ";
	else
			sBuf += "\"Cansel\"-Reject or accept \"Ok\"-temporarily? ";

	oProvider->put_RealM(Utf82Ole(sBuf));
	oProvider->put_MaySave(may_save?VARIANT_TRUE:VARIANT_FALSE);

	oProvider->FireAskEvent(SVNAuthProviderAskSslServerTrust,oProvider);

	LONG lChoice; //return value  lChoice = {0=R, 1=t, 2=p}
	oProvider->get_Choice(&lChoice);

	if (lChoice == 1)
	{
		*cred_p = (svn_auth_cred_ssl_server_trust_t *)apr_pcalloc (pool, sizeof (**cred_p));
		(*cred_p)->may_save = FALSE;
		(*cred_p)->accepted_failures = failures;
	}
	else if (may_save && lChoice == 2)
	{
		*cred_p = (svn_auth_cred_ssl_server_trust_t *)apr_pcalloc (pool, sizeof (**cred_p));
		(*cred_p)->may_save = TRUE;
		(*cred_p)->accepted_failures = failures;
	}
	else
	{
		*cred_p = NULL;
	}

  return SVN_NO_ERROR;
}


/* This implements 'svn_auth_ssl_client_cert_prompt_func_t'. */
svn_error_t *
pcom__auth_ssl_client_cert_prompt (svn_auth_cred_ssl_client_cert_t **cred_p,
                                     void *baton,
                                     const char *realm,
                                     svn_boolean_t may_save,
                                     apr_pool_t *pool)
{
	svn_auth_cred_ssl_client_cert_t *cred = (svn_auth_cred_ssl_client_cert_t *)apr_palloc (pool, sizeof(*cred));

	CComGITPtr<ISVNAuthProvider> gProvider;
	CComPtr<ISVNAuthProvider> oProvider;
	if (FAILED(gProvider.Attach(*(DWORD *)baton)))
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	if (FAILED(gProvider.CopyTo(&oProvider)))
	{
		gProvider.Detach();
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	}
	gProvider.Detach();
	oProvider->put_RealM(Utf82Ole(realm));
	oProvider->put_MaySave(may_save?VARIANT_TRUE:VARIANT_FALSE);

	oProvider->FireAskEvent(SVNAuthProviderAskSslClientCert,oProvider);
	CComBSTR bsCertFileName;
	oProvider->get_CertFileName(&bsCertFileName);

	cred->cert_file = apr_pstrdup(pool, Ole2Utf8(bsCertFileName));
	VARIANT_BOOL bMaySave;
	oProvider->get_MaySave(&bMaySave);
	cred->may_save = bMaySave==VARIANT_TRUE;
	*cred_p = cred;

	return SVN_NO_ERROR;
}


/* This implements 'svn_auth_ssl_client_cert_pw_prompt_func_t'. */
svn_error_t *
pcom__auth_ssl_client_cert_pw_prompt (
  svn_auth_cred_ssl_client_cert_pw_t **cred_p,
  void *baton,
  const char *realm,
  svn_boolean_t may_save,
  apr_pool_t *pool)
{
	svn_auth_cred_ssl_client_cert_pw_t *cred = (svn_auth_cred_ssl_client_cert_pw_t *)apr_pcalloc (pool, sizeof (*cred));

	CComGITPtr<ISVNAuthProvider> gProvider;
	CComPtr<ISVNAuthProvider> oProvider;
	if (FAILED(gProvider.Attach(*(DWORD *)baton)))
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	if (FAILED(gProvider.CopyTo(&oProvider)))
	{
		gProvider.Detach();
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Error working with GIT");
	}
	gProvider.Detach();
	oProvider->put_RealM(Utf82Ole(realm));
	oProvider->put_MaySave(may_save?VARIANT_TRUE:VARIANT_FALSE);

	oProvider->FireAskEvent(SVNAuthProviderAskSslClientCertPw,oProvider);
	CComBSTR bsPassphrase;
	oProvider->get_DefPassword(&bsPassphrase);

	cred->password = apr_pstrdup(pool, Ole2Utf8(bsPassphrase));
	VARIANT_BOOL bMaySave;
	oProvider->get_MaySave(&bMaySave);
	cred->may_save = bMaySave==VARIANT_TRUE;
	*cred_p = cred;

  return SVN_NO_ERROR;
}

STDMETHODIMP CSVNAuthProvider::Init(void)
{
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::InitEnd(VARIANT_BOOL bInterActiv, VARIANT_BOOL bNonInterActiv)
{
	// not, let's do it
	apr_status_t status;

	status = apr_initialize();
	if (status)
		return AtlReportError(__uuidof(SVNAuthProvider),_T("Cannot initialize apr m_pPool"));

	// init client m_pPool
	m_pPool = svn_pool_create(NULL);

	svn_auth_provider_object_t *provider;

	/* The whole list of registered providers */
	apr_array_header_t *providers
		= apr_array_make (m_pPool, 10, sizeof (svn_auth_provider_object_t *));

	if (bNonInterActiv == VARIANT_FALSE)
	{
		/* The main disk-caching auth providers, for both
		'username/password' creds and 'username' creds.  */
		svn_client_get_windows_simple_provider (&provider, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;
		svn_client_get_simple_provider (&provider, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;
		svn_client_get_username_provider (&provider, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;

		/* The server-cert, client-cert, and client-cert-password providers. */
		svn_client_get_ssl_server_trust_file_provider (&provider, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;
		svn_client_get_ssl_client_cert_file_provider (&provider, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;
		svn_client_get_ssl_client_cert_pw_file_provider (&provider, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;
	}

	if (bInterActiv == VARIANT_FALSE){
		m_dwCookie = m_gThis.GetCookie();
		svn_client_get_simple_prompt_provider (&provider,
			pcom__auth_simple_prompt,
			&m_dwCookie,
			2,
			m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;

		svn_client_get_username_prompt_provider (&provider,
			pcom__auth_username_prompt,
			&m_dwCookie, 
			2, 
			m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;
	        
		svn_client_get_ssl_server_trust_prompt_provider
			(&provider, pcom__auth_ssl_server_trust_prompt, &m_dwCookie, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;

		svn_client_get_ssl_client_cert_prompt_provider
			(&provider, pcom__auth_ssl_client_cert_prompt, &m_dwCookie, 2, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;

		svn_client_get_ssl_client_cert_pw_prompt_provider
			(&provider, pcom__auth_ssl_client_cert_pw_prompt, &m_dwCookie, 2, m_pPool);
		APR_ARRAY_PUSH (providers, svn_auth_provider_object_t *) = provider;
	}

	svn_auth_open (&m_ab, providers, m_pPool);

	return S_OK;
}


STDMETHODIMP CSVNAuthProvider::get_SvnAuthBaton(void ** pVal)
{
	LOCK_THIS_OBJ
	*pVal = (void*)m_ab;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::get_DefUserName(BSTR* pVal)
{
	LOCK_THIS_OBJ
	return m_bsDefUserName.CopyTo(pVal);
}

STDMETHODIMP CSVNAuthProvider::put_DefUserName(BSTR newVal)
{
	LOCK_THIS_OBJ
	m_bsDefUserName = newVal;
	svn_auth_set_parameter(m_ab, SVN_AUTH_PARAM_DEFAULT_USERNAME,apr_pstrdup (m_pPool, Ole2Utf8(m_bsDefUserName)));
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::get_DefPassword(BSTR* pVal)
{
	LOCK_THIS_OBJ
	return m_bsDefPassword.CopyTo(pVal);
}

STDMETHODIMP CSVNAuthProvider::put_DefPassword(BSTR newVal)
{
	LOCK_THIS_OBJ
	m_bsDefPassword = newVal;
	svn_auth_set_parameter(m_ab, SVN_AUTH_PARAM_DEFAULT_PASSWORD,apr_pstrdup (m_pPool, Ole2Utf8(m_bsDefPassword)));
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::get_CertFileName(BSTR* pVal)
{
	LOCK_THIS_OBJ
	return m_bsCertFileName.CopyTo(pVal);
}

STDMETHODIMP CSVNAuthProvider::put_CertFileName(BSTR newVal)
{
	LOCK_THIS_OBJ
	m_bsCertFileName = newVal;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::get_RealM(BSTR* pVal)
{
	LOCK_THIS_OBJ
	return m_bsRealM.CopyTo(pVal);
}

STDMETHODIMP CSVNAuthProvider::put_RealM(BSTR newVal)
{
	LOCK_THIS_OBJ
	m_bsRealM = newVal;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::get_MaySave(VARIANT_BOOL* pVal)
{
	if (pVal==NULL)
		return E_POINTER;
	LOCK_THIS_OBJ
	*pVal = m_bMaySave;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::put_MaySave(VARIANT_BOOL newVal)
{
	LOCK_THIS_OBJ
	m_bMaySave = newVal;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::get_Choice(LONG* pVal)
{
	if (pVal==NULL)
		return E_POINTER;
	LOCK_THIS_OBJ
	*pVal = m_lChoice;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::put_Choice(LONG newVal)
{
	LOCK_THIS_OBJ
	m_lChoice = newVal;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::get_Cookie(LONG* pVal)
{
	if (pVal==NULL)
		return E_POINTER;
	LOCK_THIS_OBJ
	*pVal = m_lCookie;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::put_Cookie(LONG newVal)
{
	LOCK_THIS_OBJ
	m_lCookie = newVal;
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::get_OptConfigDir(BSTR* pVal)
{
	LOCK_THIS_OBJ
	return m_bsOptConfigDir.CopyTo(pVal);
}

STDMETHODIMP CSVNAuthProvider::put_OptConfigDir(BSTR newVal)
{
	LOCK_THIS_OBJ
	m_bsOptConfigDir = newVal;
	svn_auth_set_parameter(m_ab, SVN_AUTH_PARAM_CONFIG_DIR,apr_pstrdup (m_pPool, Ole2Utf8(m_bsOptConfigDir)));
	return S_OK;
}

STDMETHODIMP CSVNAuthProvider::FireAskEvent(SVNAuthProviderAsk ask, ISVNAuthProvider* pProvider)
{
	Fire_AskEvent(ask,pProvider);
	return S_OK;
}

